"""
Common and constants

Supported environment variables:

::

    export DEFAULT_FETCH_DATASETS_YAHOO="pricing_yahoo,options_yahoo,
    news_yahoo"

"""

import os
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


FETCH_PRICING_YAHOO = 1000
FETCH_OPTIONS_YAHOO = 1001
FETCH_NEWS_YAHOO = 1002

DATAFEED_PRICING_YAHOO = 1100
DATAFEED_OPTIONS_YAHOO = 1101
DATAFEED_NEWS_YAHOO = 1102

DEFAULT_FETCH_DATASETS_YAHOO = [
    FETCH_PRICING_YAHOO,
    FETCH_OPTIONS_YAHOO,
    FETCH_NEWS_YAHOO
]
TIMESENSITIVE_DATASETS_YAHOO = [
    FETCH_PRICING_YAHOO,
    FETCH_OPTIONS_YAHOO,
    FETCH_NEWS_YAHOO
]

ENV_FETCH_DATASETS_YAHOO = os.getenv(
    'DEFAULT_FETCH_DATASETS_YAHOO',
    None)
FETCH_DATASETS_YAHOO = DEFAULT_FETCH_DATASETS_YAHOO
if ENV_FETCH_DATASETS_YAHOO:
    FETCH_DATASETS_YAHOO = \
        ENV_FETCH_DATASETS_YAHOO.split(',')


def get_ft_str_yahoo(
        ft_type):
    """get_ft_str_yahoo

    :param ft_type: enum fetch type value to return
                    as a string
    """
    if ft_type == FETCH_PRICING_YAHOO:
        return 'pricing_yahoo'
    elif ft_type == FETCH_OPTIONS_YAHOO:
        return 'options_yahoo'
    elif ft_type == FETCH_NEWS_YAHOO:
        return 'news_yahoo'
    else:
        return f'unsupported ft_type={ft_type}'
# end of get_ft_str_yahoo


def get_datafeed_str_yahoo(
        df_type):
    """get_datafeed_str_yahoo

    :param df_type: enum fetch type value to return
                    as a string
    """
    if df_type == DATAFEED_PRICING_YAHOO:
        return 'pricing_yahoo'
    elif df_type == DATAFEED_OPTIONS_YAHOO:
        return 'options_yahoo'
    elif df_type == DATAFEED_NEWS_YAHOO:
        return 'news_yahoo'
    else:
        return f'unsupported df_type={df_type}'
# end of get_datafeed_str_yahoo



"""
Helper for loading datasets from a file

**Supported environment variables**

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import analysis_engine.consts as ae_consts
import analysis_engine.prepare_dict_for_algo as prepare_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def load_algo_dataset_from_file(
        path_to_file,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        compress=True,
        encoding='utf-8'):
    """load_algo_dataset_from_file

    Load an algorithm-ready dataset for algorithm backtesting
    from a local file

    :param path_to_file: string - path to file holding an
        algorithm-ready dataset
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``True`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding
    """
    log.info(
        f'start: {path_to_file}')
    data_from_file = None
    file_args = 'rb'
    if not compress:
        file_args = 'r'
    with open(path_to_file, file_args) as cur_file:
        data_from_file = cur_file.read()

    if not data_from_file:
        log.error(f'missing data from file={path_to_file}')
        return None

    return prepare_utils.prepare_dict_for_algo(
        data=data_from_file,
        compress=compress,
        convert_to_dict=True,
        encoding=encoding)
# end of load_algo_dataset_from_file



"""
Helper for compressing a ``dict`` or ``pandas.DataFrame``
"""

import json
import zlib
import analysis_engine.consts as ae_consts


def compress_data(
        data,
        encoding='utf-8',
        date_format=None):
    """compress_data

    Helper for compressing ``data`` which can be
    either a ``dict`` or a ``pandas.DataFrame``
    objects with zlib.

    :param data: ``dict`` or ``pandas.DataFrame`` object
        to compress
    :param encoding: optional encoding - default is ``utf-8``
    :param date_format: optional date format - default is ``None``
    """

    converted_json = None
    if ae_consts.is_df(df=data):
        if date_format:
            converted_json = data.to_json(
                orient='records',
                date_format=date_format)
        else:
            converted_json = data.to_json(
                orient='records')
    else:
        converted_json = data

    converted_str = json.dumps(
        converted_json).encode(
            encoding)
    compressed_str = zlib.compress(converted_str, 9)

    return compressed_str
# end of compress_data



"""
Helper for creating a buy order
"""

import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_buy_order(
        ticker,
        num_owned,
        close,
        balance,
        commission,
        date,
        details,
        use_key,
        minute=None,
        shares=None,
        version=1,
        auto_fill=True,
        is_live_trading=False,
        backtest_shares_default=10,
        reason=None):
    """build_buy_order

    Create an algorithm buy order as a dictionary

    .. note:: setting the ``minute`` is required to build
        a minute-by-minute ``Trading History``

    :param ticker: ticker
    :param num_owned: integer current owned
        number of shares for this asset
    :param close: float closing price of the asset
    :param balance: float amount of available capital
    :param commission: float for commission costs
    :param date: string trade date for that row usually
        ``COMMON_DATE_FORMAT`` (``YYYY-MM-DD``)
    :param minute: optional - string with the minute that the
        order was placed. format is
        ``COMMON_TICK_DATE_FORMAT`` (``YYYY-MM-DD HH:MM:SS``)
    :param details: dictionary for full row of values to review
        all buys after the algorithm finishes.
        (usually ``row.to_json()``)
    :param use_key: string for redis and s3 publishing of the algorithm
        result dictionary as a json-serialized dictionary
    :param shares: optional - integer number of shares to buy
        if None buy max number of shares at the ``close`` with the
        available ``balance`` amount.
    :param version: optional - version tracking integer
    :param auto_fill: optional - bool for not assuming the trade
        filled (default ``True``)
    :param is_live_trading: optional - bool for filling trades
        for live trading or for backtest tuning filled
        (default ``False`` which is backtest mode)
    :param backtest_shares_default: optional - integer for
        simulating shares during a backtest even if there
        are not enough funds
        (default ``10``)
    :param reason: optional - string for recording why the algo
        decided to buy for review after the algorithm finishes
    """
    status = ae_consts.TRADE_OPEN
    s3_bucket_name = ae_consts.ALGO_BUYS_S3_BUCKET_NAME
    s3_key = use_key
    redis_key = use_key
    s3_enabled = True
    redis_enabled = True

    cost_of_trade = None
    new_shares = num_owned
    new_balance = balance
    created_date = None

    tradable_funds = balance - (2.0 * commission)

    if not is_live_trading:
        if not shares:
            shares = backtest_shares_default
        tradable_funds = (
            (shares * close) + (2.0 * commission))

    if close > 0.1 and tradable_funds > 10.0:
        can_buy_num_shares = shares
        if not can_buy_num_shares:
            can_buy_num_shares = int(tradable_funds / close)
        cost_of_trade = ae_consts.to_f(
            val=(can_buy_num_shares * close) + commission)
        if can_buy_num_shares > 0:
            if cost_of_trade > balance:
                status = ae_consts.TRADE_NOT_ENOUGH_FUNDS
            else:
                created_date = ae_utils.utc_now_str()
                if auto_fill:
                    new_shares = int(num_owned + can_buy_num_shares)
                    new_balance = ae_consts.to_f(balance - cost_of_trade)
                    status = ae_consts.TRADE_FILLED
                else:
                    new_shares = shares
                    new_balance = balance
        else:
            status = ae_consts.TRADE_NOT_ENOUGH_FUNDS
    else:
        status = ae_consts.TRADE_NOT_ENOUGH_FUNDS

    order_dict = {
        'ticker': ticker,
        'status': status,
        'balance': new_balance,
        'shares': new_shares,
        'buy_price': cost_of_trade,
        'prev_balance': balance,
        'prev_shares': num_owned,
        'close': close,
        'details': details,
        'reason': reason,
        'date': date,
        'minute': minute,
        'created': created_date,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled,
        'version': version
    }

    use_date = minute
    if not use_date:
        use_date = date

    log.debug(
        f'{ticker} {use_date} buy '
        f'{ae_consts.get_status(status=order_dict["status"])} '
        f'order={ae_consts.ppj(order_dict)}')

    return order_dict
# end of build_buy_order



"""
Base Test Case class

This class provides common functionality for most
unittests:

- turns on debugging values
- adds and removes test environment variables
- interface for building test data
"""

import os
import unittest
import analysis_engine.api_requests as api_requests


class BaseTestCase(unittest.TestCase):
    """BaseTestCase"""

    debug = False
    celery_disabled_value = None
    debug_get_pricing = None
    debug_pub_pricing = None
    backup_s3_contents = None

    def setUp(
            self):
        """setUp"""
        self.backup_s3_contents = None
        self.celery_disabled_value = None
        self.debug_get_pricing = None
        self.has_ta_lib = False
        if (os.getenv("READTHEDOCS", "") == ""
                and os.getenv("TRAVIS", "") == ""):
            self.has_ta_lib = True

        if os.getenv(
                'CELERY_DISABLED',
                'not-set') != 'not-set':
            self.celery_disabled_value = os.getenv(
                'CELERY_DISABLED',
                None)
        os.environ['CELERY_DISABLED'] = '1'
        if os.getenv(
                'DEBUG_GET_PRICING',
                'not-set') != 'not-set':
            self.debug_get_pricing = os.getenv(
                'DEBUG_GET_PRICING',
                None)
        os.environ['DEBUG_GET_PRICING'] = '1'
        if os.getenv(
                'DEBUG_PUB_PRICING',
                'not-set') != 'not-set':
            self.debug_pub_pricing = os.getenv(
                'DEBUG_PUB_PRICING',
                None)
        os.environ['DEBUG_PUB_PRICING'] = '1'
        if os.getenv(
                'TEST_S3_CONTENTS',
                'not-set') != 'not-set':
            self.backup_s3_contents = os.getenv(
                'TEST_S3_CONTENTS',
                None)
        os.environ.pop('TEST_S3_CONTENTS', None)
    # end of setUp

    def tearDown(
            self):
        """tearDown"""
        if self.celery_disabled_value:
            os.environ['CELERY_DISABLED'] = self.celery_disabled_value
        else:
            os.environ.pop('CELERY_DISABLED', None)
        if self.debug_get_pricing:
            os.environ['DEBUG_GET_PRICING'] = self.debug_get_pricing
        else:
            os.environ.pop('DEBUG_GET_PRICING', None)
        if self.debug_pub_pricing:
            os.environ['DEBUG_PUB_PRICING'] = self.debug_pub_pricing
        else:
            os.environ.pop('DEBUG_PUB_PRICING', None)
        if self.backup_s3_contents:
            os.environ['TEST_S3_CONTENTS'] = self.backup_s3_contents
        else:
            os.environ.pop('TEST_S3_CONTENTS', None)
    # end of tearDown

    def get_pricing_test_data(
            self,
            test_name=None):
        """get_pricing_test_data"""
        test_data = api_requests.build_cache_ready_pricing_dataset()
        if test_name:
            test_data['_TEST_NAME'] = test_name
        else:
            test_data['_TEST_NAME'] = 'not-set'
        return test_data
    # end of get_pricing_test_data

# end of BaseTestCase



"""
Load an ``Trading History`` dataset from file, s3 -
redis coming soon

Supported Datasets:

- ``SA_DATASET_TYPE_TRADING_HISTORY`` - trading history datasets

"""

import os
import analysis_engine.consts as consts
import analysis_engine.load_history_dataset_from_file as file_utils
import analysis_engine.load_history_dataset_from_s3 as s3_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def load_history_dataset(
        history_dataset=None,
        dataset_type=None,
        serialize_datasets=None,
        path_to_file=None,
        compress=None,
        encoding='utf-8',
        convert_to_dict=False,
        redis_enabled=None,
        redis_key=None,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_serializer='json',
        redis_encoding='utf-8',
        s3_enabled=None,
        s3_key=None,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=None,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        verbose=False):
    """load_history_dataset

    Load a ``Trading History`` Dataset from file, s3 - note
    redis is not supported yet

    :param history_dataset: optional - already loaded history dataset
    :param dataset_type: optional - dataset type
        (default is ``analysis_engine.consts.SA_DATASET_TYPE_TRADING_HISTORY``)
    :param path_to_file: optional - path to a trading history dataset
        in a file
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``True`` and uses ``zlib`` for compression)
    :param encoding: optional - string for data encoding
    :param convert_to_dict: optional - boolean flag for decoding
        as a dictionary during prepare

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``analysis_engine.consts.ENABLED_REDIS_PUBLISH``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``analysis_engine.consts.REDIS_ADDRESS``)
    :param redis_db: Redis db to use
        (default is ``analysis_engine.consts.REDIS_DB``)
    :param redis_password: optional - Redis password
        (default is ``analysis_engine.consts.REDIS_PASSWORD``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis
        (default is ``json``)
    :param redis_encoding: format of the encoded key in redis
        (default is ``utf-8``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``analysis_engine.consts.ENABLED_S3_UPLOAD``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``analysis_engine.consts.S3_ADDRESS``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``analysis_engine.consts.S3_BUCKET``) which should be
        viewable on a browser:
        http://localhost:9000/minio/
    :param s3_access_key: S3 Access key
        (default is ``analysis_engine.consts.S3_ACCESS_KEY``)
    :param s3_secret_key: S3 Secret key
        (default is ``analysis_engine.consts.S3_SECRET_KEY``)
    :param s3_region_name: S3 region name
        (default is ``analysis_engine.consts.S3_REGION_NAME``)
    :param s3_secure: Transmit using tls encryption
        (default is ``analysis_engine.consts.S3_SECURE``)

    **(Optional) Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed

    Additonal arguments

    :param verbose: optional - bool for increasing
        logging
    """

    if not dataset_type:
        dataset_type = consts.SA_DATASET_TYPE_TRADING_HISTORY
    if not serialize_datasets:
        serialize_datasets = consts.DEFAULT_SERIALIZED_DATASETS
    if not redis_enabled:
        redis_enabled = consts.ENABLED_REDIS_PUBLISH
    if not redis_address:
        redis_address = consts.REDIS_ADDRESS
    if not redis_db:
        redis_db = consts.REDIS_DB
    if not redis_password:
        redis_password = consts.REDIS_PASSWORD
    if not s3_enabled:
        s3_enabled = consts.ENABLED_S3_UPLOAD
    if not s3_address:
        s3_address = consts.S3_ADDRESS
    if not s3_bucket:
        s3_bucket = consts.S3_BUCKET
    if not s3_access_key:
        s3_access_key = consts.S3_ACCESS_KEY
    if not s3_secret_key:
        s3_secret_key = consts.S3_SECRET_KEY
    if not s3_region_name:
        s3_region_name = consts.S3_REGION_NAME
    if not s3_secure:
        s3_secure = consts.S3_SECURE
    if compress is None:
        compress = consts.ALGO_HISTORY_COMPRESS

    use_ds = history_dataset
    if not use_ds:
        log.info(
            f'loading {consts.get_status(status=dataset_type)} from '
            f'file={path_to_file} s3={s3_key} redis={redis_key}')
    # load if not created

    supported_type = False
    if dataset_type == consts.SA_DATASET_TYPE_TRADING_HISTORY:
        supported_type = True
        if (path_to_file and
                not use_ds):
            if not os.path.exists(path_to_file):
                log.error(f'missing file: {path_to_file}')
            use_ds = file_utils.load_history_dataset_from_file(
                path_to_file=path_to_file,
                compress=compress,
                encoding=redis_encoding)
        elif (s3_key and
                not use_ds):
            use_ds = s3_utils.load_history_dataset_from_s3(
                s3_key=s3_key,
                s3_address=s3_address,
                s3_bucket=s3_bucket,
                s3_access_key=s3_access_key,
                s3_secret_key=s3_secret_key,
                s3_region_name=s3_region_name,
                s3_secure=s3_secure,
                compress=compress,
                encoding=redis_encoding,
                convert_to_dict=convert_to_dict,
                serialize_datasets=serialize_datasets)
    else:
        supported_type = False
        use_ds = None
        log.error(
            f'loading {dataset_type} from file={path_to_file} '
            f's3={s3_key} redis={redis_key}')
    # load if not created

    if not use_ds and supported_type:
        log.error(
            f'unable to load a dataset from file={path_to_file} '
            f's3={s3_key} redis={redis_key}')

    return use_ds
# end of load_history_dataset



"""
Build option spread pricing details
"""

import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_option_spread_details(
        trade_type,
        spread_type,
        option_type,
        close,
        num_contracts,
        low_strike,
        low_ask,
        low_bid,
        high_strike,
        high_ask,
        high_bid):
    """build_option_spread_details

    Calculate pricing information for supported spreads
    including ``max loss``, ``max profit``, and ``mid price`` (break
    even coming soon)

    :param trade_type: entry (``TRADE_ENTRY``) or
        exit (``TRADE_EXIT``) of a spread position
    :param spread_type: vertical bull (``SPREAD_VERTICAL_BULL``)
        and vertical bear (``SPREAD_VERTICAL_BEAR``)
        are the only supported calculations for now
    :param option_type: call (``OPTION_CALL``) or put
        (``OPTION_PUT``)
    :param close: closing price of the underlying
        asset
    :param num_contracts: integer number of contracts
    :param low_strike: float - strike for
        the low leg of the spread
    :param low_ask: float - ask price for
        the low leg of the spread
    :param low_bid: float - bid price for
        the low leg of the spread
    :param high_strike: float - strike  for
        the high leg of the spread
    :param high_ask: float - ask price for
        the high leg of the spread
    :param high_bid: float - bid price for
        the high leg of the spread
    """

    details = {
        'status': ae_consts.NOT_RUN,
        'trade_type': trade_type,
        'spread_type': spread_type,
        'option_type': option_type,
        'num_contracts': num_contracts,
        'low_strike': low_strike,
        'low_bid': low_bid,
        'low_ask': low_ask,
        'high_strike': high_strike,
        'high_bid': high_bid,
        'high_ask': high_ask,
        'cost': None,
        'revenue': None,
        'low_bidask_mid': None,
        'high_bidask_mid': None,
        'mid_price': None,
        'nat_price': None,
        'strike_width': None,
        'break_even': None,
        'max_loss': None,
        'max_profit': None,
        'spread_id': None
    }

    low_distance = int(close) - low_strike
    high_distance = high_strike - int(close)
    details['strike_width'] = ae_consts.to_f(
        high_strike - low_strike)
    details['spread_id'] = (
        f'T_{trade_type}_S_{spread_type}_O_{option_type}_low_{low_distance}'
        f'_high_{high_distance}_w_{details["strike_width"]}')
    details['low_bidask_mid'] = ae_consts.to_f(low_bid + low_ask / 2.0)
    details['high_bidask_mid'] = ae_consts.to_f(high_bid + high_ask / 2.0)
    details['mid_price'] = ae_consts.to_f(abs(
        details['low_bidask_mid'] - details['high_bidask_mid']))
    details['nat_price'] = ae_consts.to_f(abs(
        details['low_bidask_mid'] - details['high_bidask_mid']))

    cost_of_contracts_at_mid_price = None
    revenue_of_contracts_at_mid_price = None

    if trade_type == ae_consts.TRADE_ENTRY:
        cost_of_contracts_at_mid_price = ae_consts.to_f(
            100.0 * num_contracts * details['mid_price'])
        revenue_of_contracts_at_mid_price = ae_consts.to_f(
            100.0 * num_contracts * (
                details['strike_width'] - details['mid_price']))
        if spread_type == ae_consts.SPREAD_VERTICAL_BULL:
            if option_type == ae_consts.OPTION_CALL:  # debit spread
                details['max_loss'] = cost_of_contracts_at_mid_price
                details['max_profit'] = revenue_of_contracts_at_mid_price
            else:
                details['max_loss'] = cost_of_contracts_at_mid_price
                details['max_profit'] = revenue_of_contracts_at_mid_price
        else:  # bear
            if option_type == ae_consts.OPTION_CALL:  # debit spread
                details['max_loss'] = cost_of_contracts_at_mid_price
                details['max_profit'] = revenue_of_contracts_at_mid_price
            else:
                details['max_loss'] = cost_of_contracts_at_mid_price
                details['max_profit'] = revenue_of_contracts_at_mid_price

    else:  # trade exit calculations:
        revenue_of_contracts_at_mid_price = ae_consts.to_f(
            100.0 * num_contracts * details['mid_price'])
        cost_of_contracts_at_mid_price = ae_consts.to_f(
            100.0 * num_contracts * (
                details['strike_width'] - details['mid_price']))
        if spread_type == ae_consts.SPREAD_VERTICAL_BULL:
            if option_type == ae_consts.OPTION_CALL:  # credit spread
                details['max_profit'] = revenue_of_contracts_at_mid_price
                details['max_loss'] = cost_of_contracts_at_mid_price
            else:
                details['max_profit'] = revenue_of_contracts_at_mid_price
                details['max_loss'] = cost_of_contracts_at_mid_price
        else:  # bear
            if option_type == ae_consts.OPTION_CALL:  # credit spread
                details['max_profit'] = revenue_of_contracts_at_mid_price
                details['max_loss'] = cost_of_contracts_at_mid_price
            else:
                details['max_profit'] = revenue_of_contracts_at_mid_price
                details['max_loss'] = cost_of_contracts_at_mid_price
    # end of supported types of spreads

    details['cost'] = cost_of_contracts_at_mid_price
    details['revenue'] = revenue_of_contracts_at_mid_price

    log.debug(
        f'type={ae_consts.get_status(status=trade_type)} '
        f'spread={ae_consts.get_status(status=spread_type)} '
        f'option={ae_consts.get_status(status=option_type)} '
        f'close={close} spread_id={details["spread_id"]} '
        f'revenue={revenue_of_contracts_at_mid_price} '
        f'cost={cost_of_contracts_at_mid_price} '
        f'mid={details["mid_price"]} width={details["strike_width"]} '
        f'max_profit={details["max_profit"]} max_loss={details["max_loss"]}')

    return details
# end of build_option_spread_details



"""
Get Pricing using most of the pieces of example from:
https://github.com/timkpaine/hedgedata/blob/master/scripts/fetch_data.py

This will fetch various pricing, financials, earnings,
dividends, news and other data from
`IEX <https://iexcloud.io/>`__.

Data Attribution
================

IEX Real-Time Price
===================

- Cite IEX using the following text and link:
  "Data provided by [IEX](https://iextrading.com/developer/docs/)."
- Provide a link to
  https://iextrading.com/api-exhibit-a in your terms of service.
- Additionally, if you display our TOPS price data,
  cite "IEX Real-Time Price" near the price.

"""

import os
import os.path
import pandas as pd
import analysis_engine.iex.consts
from hedgedata.backfill import whichBackfill
from hedgedata.data import FIELDS
from hedgedata.distributor import Distributer
from spylunking.log.setup_logging import build_colorized_logger
# from hedgedata import sp500_constituents


log = build_colorized_logger(
    name=__name__)


_DISTRIBUTOR = Distributer.default()


def backfillData(
        distributor,
        symbols,
        fields,
        output='cache'):
    """backfillData

    :param distributor: distributor object from:
                        https://github.com/timkpaine/hedgedata
    :param symbols: list of symbols to iterate
    :param fields: path to files
    :param output: output dir
    """
    if not os.path.exists(output):
        os.makedirs(output)

    for field in fields:
        if os.path.exists(os.path.join(output, field) + '.csv'):
            data_orig = pd.read_csv(os.path.join(output, field) + '.csv')
            for k in (
                    'date',
                    'datetime',
                    'reportDate',
                    'EPSReportDate',
                    'exDate'):
                if k in data_orig.columns:
                    data_orig[k] = pd.to_datetime(data_orig[k])

        else:
            data_orig = pd.DataFrame()

        for symbol, data in whichBackfill(field)(_DISTRIBUTOR, symbols):
            if data.empty:
                log.critical('Skipping %s for %s' % (symbol, field))
                continue

            log.critical('Filling %s for %s' % (symbol, field))
            data.reset_index(inplace=True)

            if field == 'PEERS':
                data = data[['peer']]

            data['KEY'] = symbol
            data_orig = pd.concat(
                [data_orig, data]) if not data_orig.empty else data

        if not data_orig.empty:
            data_orig.set_index(
                analysis_engine.iex.consts.get_default_fields(
                    field),
                inplace=True)
            data_orig[
                ~data_orig.index.duplicated(
                    keep='first')].to_csv(
                        os.path.join(output, field) + '.csv')


if __name__ == '__main__':
    syms = [
        'SPY',
        'AMZN',
        'TLSA',
        'NFLX'
    ]
    backfillData(
        symbols=syms,
        fields=FIELDS,
        output='./cache')



"""
Mock TA-Lib objects
"""

import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def MockWILLRBuy(
        high=None,
        low=None,
        close=None,
        timeperiod=None):
    """MockWILLRBuy

    build a mock wiliams r object that will report
    an ``buy`` value to test indicators without
    having talib installed

    :param high: list of highs
    :param low: list of lows
    :param close: list of closes
    :param timeperiod: integer number of values
        in ``high``, ``low`` and ``close``
    """
    log.warn('mock - MockTALib.WILLR - BUY')
    retval = []
    for h in high:
        retval.append(None)
    retval[-1] = -99.9
    return retval
# end of MockWILLRBuy


def MockWILLRSell(
        high=None,
        low=None,
        close=None,
        timeperiod=None):
    """MockWILLRSell

    build a mock wiliams r object that will report
    an ``sell`` value to test indicators without
    having talib installed

    :param high: list of highs
    :param low: list of lows
    :param close: list of closes
    :param timeperiod: integer number of values
        in ``high``, ``low`` and ``close``
    """
    log.warn('mock - MockTALib.WILLR - SELL')
    retval = []
    for h in high:
        retval.append(None)
    retval[-1] = -1.0
    return retval
# end of MockWILLRSell


def MockWILLRIgnore(
        high=None,
        low=None,
        close=None,
        timeperiod=None):
    """MockWILLRIgnore

    build a mock wiliams r object that will report
    an ``ignore`` value to test indicators without
    having talib installed

    :param high: list of highs
    :param low: list of lows
    :param close: list of closes
    :param timeperiod: integer number of values
        in ``high``, ``low`` and ``close``
    """
    log.warn('mock - MockTALib.WILLR - IGNORE')
    retval = []
    for h in high:
        retval.append(None)
    retval[-1] = -50.0
    return retval
# end of MockWILLRIgnore



"""
**Publish Pricing Data Task**

Publish new stock data to external services and systems
(redis and s3) provided the system(s) are running and enabled.

- redis - using `redis-py <https://github.com/andymccurdy/redis-py>`__
- s3 - using boto3

**Sample work_dict request for this method**

`analysis_engine.api_requests.publish_pricing_update <https://
github.com/AlgoTraders/stock-analysis-engine/blob/master/
analysis_engine/api_requests.py#L344>`__

::

    work_request = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        'strike': use_strike,
        'contract': contract_type,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'data': use_data
    }

.. tip:: This task uses the `analysis_engine.work_tasks.
    custom_task.CustomTask class <https://github.com/A
    lgoTraders/stock-analysis-engine/blob/master/anal
    ysis_engine/work_tasks/custom_task.py>`__ for
    task event handling.

**Supported Environment Variables**

::

    export DEBUG_RESULTS=1

"""

import boto3
import redis
import json
import zlib
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import analysis_engine.get_task_results as get_task_results
import analysis_engine.work_tasks.custom_task as custom_task
import analysis_engine.set_data_in_redis_key as redis_set
import celery.task as celery_task
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


@celery_task(
    bind=True,
    base=custom_task.CustomTask,
    queue='publish_pricing_update')
def publish_pricing_update(
        self,
        work_dict):
    """publish_pricing_update

    Publish Ticker Data to S3 and Redis

    - prices - turn off with ``work_dict.get_pricing = False``
    - news - turn off with ``work_dict.get_news = False``
    - options - turn off with ``work_dict.get_options = False``

    :param work_dict: dictionary for key/values
    """

    label = 'publish_pricing'

    log.debug(f'task - {label} - start')

    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    rec = {
        'ticker': None,
        'ticker_id': None,
        's3_enabled': False,
        'redis_enabled': False,
        's3_bucket': None,
        's3_key': None,
        'redis_key': None,
        'updated': None
    }
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:
        ticker = work_dict.get(
            'ticker',
            ae_consts.TICKER)
        ticker_id = int(work_dict.get(
            'ticker_id',
            ae_consts.TICKER_ID))

        if not ticker:
            res = build_result.build_result(
                status=ae_consts.ERR,
                err='missing ticker',
                rec=rec)
            return res

        label = work_dict.get(
            'label',
            label)
        s3_key = work_dict.get(
            's3_key',
            None)
        s3_bucket_name = work_dict.get(
            's3_bucket',
            'pricing')
        redis_key = work_dict.get(
            'redis_key',
            None)
        data = work_dict.get(
            'data',
            None)
        updated = work_dict.get(
            'updated',
            None)
        enable_s3_upload = work_dict.get(
            's3_enabled',
            ae_consts.ENABLED_S3_UPLOAD)
        enable_redis_publish = work_dict.get(
            'redis_enabled',
            ae_consts.ENABLED_REDIS_PUBLISH)
        serializer = work_dict.get(
            'serializer',
            'json')
        encoding = work_dict.get(
            'encoding',
            'utf-8')

        rec['ticker'] = ticker
        rec['ticker_id'] = ticker_id
        rec['s3_bucket'] = s3_bucket_name
        rec['s3_key'] = s3_key
        rec['redis_key'] = redis_key
        rec['updated'] = updated
        rec['s3_enabled'] = enable_s3_upload
        rec['redis_enabled'] = enable_redis_publish

        if (
                enable_s3_upload and
                s3_bucket_name != 'MISSING_AN_S3_BUCKET'):
            access_key = work_dict.get(
                's3_access_key',
                ae_consts.S3_ACCESS_KEY)
            secret_key = work_dict.get(
                's3_secret_key',
                ae_consts.S3_SECRET_KEY)
            region_name = work_dict.get(
                's3_region_name',
                ae_consts.S3_REGION_NAME)
            service_address = work_dict.get(
                's3_address',
                ae_consts.S3_ADDRESS)
            secure = work_dict.get(
                's3_secure',
                ae_consts.S3_SECURE) == '1'

            endpoint_url = f'http{"s" if secure else ""}://{service_address}'

            log.debug(
                f'{label} building s3 endpoint_url={endpoint_url} '
                f'region={region_name}')

            s3 = boto3.resource(
                's3',
                endpoint_url=endpoint_url,
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                region_name=region_name,
                config=boto3.session.Config(
                    signature_version='s3v4')
            )

            try:
                log.debug(f'{label} checking bucket={s3_bucket_name} exists')
                if s3.Bucket(s3_bucket_name) not in s3.buckets.all():
                    log.debug(f'{label} creating bucket={s3_bucket_name}')
                    s3.create_bucket(
                        Bucket=s3_bucket_name)
            except Exception as e:
                log.debug(
                    f'{label} failed creating bucket={s3_bucket_name} '
                    f'with ex={e}')
            # end of try/ex for creating bucket

            try:
                log.debug(
                    f'{label} uploading to s3={s3_bucket_name}/{s3_key} '
                    f'updated={updated}')
                s3.Bucket(s3_bucket_name).put_object(
                    Key=s3_key,
                    Body=json.dumps(data).encode(encoding))
            except Exception as e:
                log.error(
                    f'{label} failed uploading bucket={s3_bucket_name} '
                    f'key={s3_key} ex={e}')
            # end of try/ex for creating bucket
        else:
            log.debug(
                f'{label} SKIP S3 upload bucket={s3_bucket_name} '
                f'key={s3_key}')
        # end of if enable_s3_upload

        if enable_redis_publish:
            redis_address = work_dict.get(
                'redis_address',
                ae_consts.REDIS_ADDRESS)
            redis_key = work_dict.get(
                'redis_key',
                ae_consts.REDIS_KEY)
            redis_password = work_dict.get(
                'redis_password',
                ae_consts.REDIS_PASSWORD)
            redis_db = work_dict.get(
                'redis_db',
                ae_consts.REDIS_DB)
            redis_expire = None
            if 'redis_expire' in work_dict:
                redis_expire = work_dict.get(
                    'redis_expire',
                    ae_consts.REDIS_EXPIRE)
            log.debug(
                f'redis enabled address={redis_address}@{redis_db} '
                f'key={redis_key}')
            redis_host = None
            redis_port = None
            try:
                redis_host = redis_address.split(':')[0]
                redis_port = redis_address.split(':')[1]
            except Exception as c:
                err = (
                    f'{label} failed parsing redis_address={redis_address} '
                    f'with ex={c} '
                    'please set one with the format: '
                    '<hostname>:<port>')
                log.critical(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=rec)
                return res
            # end of checking that redis_address is valid

            try:
                log.debug(
                    f'{label} publishing redis={redis_host}:{redis_port} '
                    f'db={redis_db} key={redis_key} '
                    f'updated={updated} expire={redis_expire}')

                rc = redis.Redis(
                    host=redis_host,
                    port=redis_port,
                    password=redis_password,
                    db=redis_db)

                already_compressed = False
                uses_data = data
                try:
                    uses_data = zlib.compress(json.dumps(data).encode(
                        encoding),
                        9)
                    already_compressed = True
                except Exception as p:
                    log.critical(
                        'failed to compress dataset for '
                        f'redis_key={redis_key} with ex={p}')

                redis_set_res = redis_set.set_data_in_redis_key(
                    label=label,
                    client=rc,
                    key=redis_key,
                    data=uses_data,
                    already_compressed=already_compressed,
                    serializer=serializer,
                    encoding=encoding,
                    expire=redis_expire,
                    px=None,
                    nx=False,
                    xx=False)

                log.debug(
                    f'{label} redis_set '
                    f'status={ae_consts.get_status(redis_set_res["status"])} '
                    f'err={redis_set_res["err"]}')

            except Exception as e:
                log.error(
                    f'{label} failed - redis publish to '
                    f'key={redis_key} ex={e}')
            # end of try/ex for creating bucket
        else:
            log.debug(f'{label} SKIP REDIS publish key={redis_key}')
        # end of if enable_redis_publish

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                'failed - publish_pricing_update '
                f'dict={work_dict} with ex={e}'),
            rec=rec)
        log.error(
            f'{label} - {res["err"]}')
    # end of try/ex

    log.debug(
        f'task - publish_pricing_update done - {label} - '
        f'status={ae_consts.get_status(res["status"])}')

    return get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of publish_pricing_update


def run_publish_pricing_update(
        work_dict):
    """run_publish_pricing_update

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    label = work_dict.get(
        'label',
        '')

    log.debug(f'run_publish_pricing_update - {label} - start')

    response = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    # allow running without celery
    if ae_consts.is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = publish_pricing_update(
            work_dict=work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ae_consts.ppj(response)
                except Exception:
                    response_details = response
                log.debug(f'{label} task result={response_details}')
        else:
            log.error(
                f'{label} celery was disabled but the task={response} '
                'did not return anything')
        # end of if response
    else:
        task_res = publish_pricing_update.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
            log.debug(
                f'run_publish_pricing_update - {label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]} rec={response["rec"]}')
        else:
            log.debug(
                f'run_publish_pricing_update - {label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]}')
    else:
        log.debug(f'run_publish_pricing_update - {label} - done no response')
    # end of if/else response

    return response
# end of run_publish_pricing_update



"""
Build an authenticated url for IEX Cloud
"""


def build_auth_url(
        url,
        token=None):
    """build_auth_url

    Helper for constructing authenticated IEX urls
    using an ``IEX Publishable Token`` with a valid
    `IEX Cloud Beta Account <https://
    iexcloud.io/cloud-login#/register/>`__

    This will return a string with the token as a query
    parameter on the HTTP url

    :param url: initial url to make authenticated
    :param token: optional - string ``IEX Publishable Token``
        (defaults to ``IEX_TOKEN`` environment variable or
        ``None``)
    """
    if token:
        return (
            f'{url}?token={token}')
    else:
        return url
# end of build_auth_url



"""
Build a deep neural network for regression predictions
"""

import json
import spylunking.log.setup_logging as log_utils
import keras.models as keras_models
import keras.layers as keras_layers

log = log_utils.build_colorized_logger(
    name=__name__)


def build_regression_dnn(
        num_features,
        compile_config,
        model_json=None,
        model_config=None):
    """build_regression_dnn

    :param num_features: input_dim for the number of
                         features in the data
    :param compile_config: dictionary of compile options
    :param model_json: keras model json to build the model
    :param model_config: optional dictionary for model
    """

    model = None
    num_layers = 0

    if model_json:
        log.info(
            f'loading from model_json={model_json}')
        model = keras_models.model_from_json(
            json.dumps(model_json))
    elif model_config:
        model = keras_models.Sequential()
        log.info(
            f'building '
            f'dnn num_features={num_features} '
            f'model_config={model_config}')
        for idx, node in enumerate(model_config['layers']):
            layer_type = node.get(
                'layer_type',
                'dense').lower()
            if layer_type == 'dense':
                if num_layers == 0:
                    model.add(
                        keras_layers.Dense(
                            int(node['num_neurons']),
                            input_dim=num_features,
                            kernel_initializer=node['init'],
                            activation=node['activation']))
                else:
                    model.add(
                        keras_layers.Dense(
                            int(node['num_neurons']),
                            kernel_initializer=node['init'],
                            activation=node['activation']))
            else:
                if layer_type == 'dropout':
                    model.add(
                        keras_models.Dropout(
                            float(node['rate'])))
            # end of supported model types
            num_layers += 1
        # end of all layers
    else:
        # https://machinelearningmastery.com/regression-tutorial-keras-deep-learning-library-python/  # noqa
        log.info(
            f'default dnn num_features={num_features}')
        model.add(
            keras_layers.Dense(
                8,
                input_dim=num_features,
                kernel_initializer='normal',
                activation='relu'))
        model.add(
            keras_layers.Dense(
                6,
                kernel_initializer='normal',
                activation='relu'))
        model.add(
            keras_layers.Dense(
                1,
                kernel_initializer='normal'))
    # end of building a regression dnn

    # if model was defined
    if model:
        log.info(
            f'compiling={compile_config}')
        # compile the model
        loss = compile_config.get(
            'loss',
            'mse')
        optimizer = compile_config.get(
            'optimizer',
            'adam')
        metrics = compile_config.get(
            'metrics',
            [
                'mse',
                'mae',
                'mape',
                'cosine'
            ])
        model.compile(
            loss=loss,
            optimizer=optimizer,
            metrics=metrics)
    else:
        log.error(
            f'failed building regression model={model}')
    # if could compile model

    return model
# end of build_regression_dnn



"""
Common Fetch for any supported Get from IEX using HTTP

Supported environment variables:

::

    # debug the fetch routines with:
    export DEBUG_IEX_DATA=1

"""

import os
import datetime
import copy
import analysis_engine.consts as ae_consts
import analysis_engine.iex.consts as iex_consts
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import analysis_engine.iex.fetch_data as iex_fetch_data
import analysis_engine.work_tasks.publish_pricing_update as publisher
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_data_from_iex(
        work_dict):
    """get_data_from_iex

    Get data from IEX - this requires an account

    :param work_dict: request dictionary
    """
    label = 'get_data_from_iex'

    log.debug(
        f'task - {label} - start '
        f'work_dict={work_dict}')

    rec = {
        'data': None,
        'updated': None
    }
    res = {
        'status': ae_consts.NOT_RUN,
        'err': None,
        'rec': rec
    }

    ticker = None
    field = None
    ft_type = None

    try:

        ticker = work_dict.get('ticker', ae_consts.TICKER)
        field = work_dict.get('field', 'daily')
        ft_type = work_dict.get('ft_type', None)
        ft_str = str(ft_type).lower()
        label = work_dict.get('label', label)
        orient = work_dict.get('orient', 'records')
        backfill_date = work_dict.get('backfill_date', None)

        iex_req = None
        if ft_type == iex_consts.FETCH_DAILY or ft_str == 'daily':
            ft_type == iex_consts.FETCH_DAILY
            iex_req = api_requests.build_iex_fetch_daily_request(
                label=label)
        elif ft_type == iex_consts.FETCH_MINUTE or ft_str == 'minute':
            ft_type == iex_consts.FETCH_MINUTE
            iex_req = api_requests.build_iex_fetch_minute_request(
                label=label)
        elif ft_type == iex_consts.FETCH_QUOTE or ft_str == 'quote':
            ft_type == iex_consts.FETCH_QUOTE
            iex_req = api_requests.build_iex_fetch_quote_request(
                label=label)
        elif ft_type == iex_consts.FETCH_STATS or ft_str == 'stats':
            ft_type == iex_consts.FETCH_STATS
            iex_req = api_requests.build_iex_fetch_stats_request(
                label=label)
        elif ft_type == iex_consts.FETCH_PEERS or ft_str == 'peers':
            ft_type == iex_consts.FETCH_PEERS
            iex_req = api_requests.build_iex_fetch_peers_request(
                label=label)
        elif ft_type == iex_consts.FETCH_NEWS or ft_str == 'news':
            ft_type == iex_consts.FETCH_NEWS
            iex_req = api_requests.build_iex_fetch_news_request(
                label=label)
        elif ft_type == iex_consts.FETCH_FINANCIALS or ft_str == 'financials':
            ft_type == iex_consts.FETCH_FINANCIALS
            iex_req = api_requests.build_iex_fetch_financials_request(
                label=label)
        elif ft_type == iex_consts.FETCH_EARNINGS or ft_str == 'earnings':
            ft_type == iex_consts.FETCH_EARNINGS
            iex_req = api_requests.build_iex_fetch_earnings_request(
                label=label)
        elif ft_type == iex_consts.FETCH_DIVIDENDS or ft_str == 'dividends':
            ft_type == iex_consts.FETCH_DIVIDENDS
            iex_req = api_requests.build_iex_fetch_dividends_request(
                label=label)
        elif ft_type == iex_consts.FETCH_COMPANY or ft_str == 'company':
            ft_type == iex_consts.FETCH_COMPANY
            iex_req = api_requests.build_iex_fetch_company_request(
                label=label)
        else:
            log.error(
                f'{label} - unsupported ft_type={ft_type} '
                f'ft_str={ft_str} ticker={ticker}')
            raise NotImplementedError
        # if supported fetch request type

        iex_req['ticker'] = ticker
        clone_keys = [
            'ticker',
            's3_address',
            's3_bucket',
            's3_key',
            'redis_address',
            'redis_db',
            'redis_password',
            'redis_key'
        ]

        for k in clone_keys:
            if k in iex_req:
                iex_req[k] = work_dict.get(k, f'{k}-missing-in-{label}')
        # end of cloning keys

        if not iex_req:
            err = (
                f'{label} - ticker={ticker} '
                f'did not build an IEX request '
                f'for work={work_dict}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            log.debug(
                f'{label} - ticker={ticker} '
                f'field={field} '
                f'orient={orient} fetch')
        # if invalid iex request

        df = None
        try:
            if 'from' in work_dict:
                iex_req['from'] = datetime.datetime.strptime(
                    '%Y-%m-%d %H:%M:%S',
                    work_dict['from'])
            if backfill_date:
                iex_req['backfill_date'] = backfill_date
                iex_req['redis_key'] = (
                    f'{ticker}_{backfill_date}_{field}')
                iex_req['s3_key'] = (
                    f'{ticker}_{backfill_date}_{field}')

            if os.getenv('SHOW_SUCCESS', '0') == '1':
                log.info(
                    f'fetching IEX {field} req={iex_req}')
            else:
                log.debug(
                    f'fetching IEX {field} req={iex_req}')

            df = iex_fetch_data.fetch_data(
                work_dict=iex_req,
                fetch_type=ft_type)
            rec['data'] = df.to_json(
                orient=orient,
                date_format='iso')
            rec['updated'] = datetime.datetime.utcnow().strftime(
                '%Y-%m-%d %H:%M:%S')
        except Exception as f:
            log.error(
                f'{label} - ticker={ticker} field={ft_type} '
                f'failed fetch_data '
                f'with ex={f}')
        # end of try/ex

        if ae_consts.ev('DEBUG_IEX_DATA', '0') == '1':
            log.debug(
                f'{label} ticker={ticker} '
                f'field={field} data={rec["data"]} to_json')
        else:
            log.debug(
                f'{label} ticker={ticker} field={field} to_json')
        # end of if/else found data

        upload_and_cache_req = copy.deepcopy(iex_req)
        upload_and_cache_req['celery_disabled'] = True
        upload_and_cache_req['data'] = rec['data']
        if not upload_and_cache_req['data']:
            upload_and_cache_req['data'] = '{}'
        use_field = field
        if use_field == 'news':
            use_field = 'news1'
        if 'redis_key' in work_dict:
            rk = work_dict.get('redis_key', iex_req['redis_key'])
            if backfill_date:
                rk = f'{ticker}_{backfill_date}'
            upload_and_cache_req['redis_key'] = (
                f'{rk}_{use_field}')
        if 's3_key' in work_dict:
            sk = work_dict.get('s3_key', iex_req['s3_key'])
            if backfill_date:
                sk = f'{ticker}_{backfill_date}'
            upload_and_cache_req['s3_key'] = (
                f'{sk}_{use_field}')

        try:
            update_res = publisher.run_publish_pricing_update(
                work_dict=upload_and_cache_req)
            update_status = update_res.get(
                'status',
                ae_consts.NOT_SET)
            log.debug(
                f'{label} publish update '
                f'status={ae_consts.get_status(status=update_status)} '
                f'data={update_res}')
        except Exception:
            err = (
                f'{label} - failed to upload iex '
                f'data={upload_and_cache_req} to '
                f'to s3_key={upload_and_cache_req["s3_key"]} '
                f'and redis_key={upload_and_cache_req["redis_key"]}')
            log.error(err)
        # end of try/ex to upload and cache

        if not rec['data']:
            log.debug(
                f'{label} - ticker={ticker} no IEX data '
                f'field={field} to publish')
        # end of if/else

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                f'failed - get_data_from_iex '
                f'dict={work_dict} with ex={e}'),
            rec=rec)
    # end of try/ex

    log.debug(
        f'task - get_data_from_iex done - '
        f'{label} - '
        f'status={ae_consts.get_status(res["status"])} '
        f'err={res["err"]}')

    return res
# end of get_data_from_iex



"""
Helper for loading a ``Trading History`` dataset
"""

import json
import zlib
import pandas as pd
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def prepare_history_dataset(
        data,
        compress=False,
        encoding='utf-8',
        convert_to_dict=False,
        include_keys=None,
        ignore_keys=None,
        convert_to_dates=None,
        verbose=False):
    """prepare_history_dataset

    Load a ``Trading History`` dataset into a dictionary
    with a ``pd.DataFrame`` for the trading history record
    list

    :param data: string holding contents of a ``Trading History``
        from a file, s3 key or redis-key
    :param compress: optional - boolean flag for decompressing
        the contents of the ``data`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param convert_to_dict: optional - bool for s3 use ``False``
        and for files use ``True``
    :param encoding: optional - string for data encoding
    :param include_keys: optional - list of string keys
        to include before from the dataset
        .. note:: tickers are automatically included in the ``pd.DataFrame``
    :param ignore_keys: optional - list of string keys
        to remove before building the ``pd.DataFrame``
    :param convert_to_dates: optional - list of string keys
        to convert to datetime before building the ``pd.DataFrame``
    :param verbose: optional - bool show the logs
        (default is ``False``)
    """
    if verbose:
        log.debug('start')
    use_data = None
    parsed_data = None
    data_as_dict = None

    if compress:
        if verbose:
            log.debug('decompressing')
        parsed_data = zlib.decompress(
            data).decode(
                encoding)
    else:
        parsed_data = data

    if not parsed_data:
        log.error('failed parsing')
        return None

    if verbose:
        log.debug('loading as dict')
    use_data = {}
    if convert_to_dict:
        try:
            data_as_dict = json.loads(parsed_data)
        except Exception as e:
            if (
                    'the JSON object must be str, bytes or '
                    'bytearray, not') in str(e):
                log.critical(
                    f'failed decoding json for string - double '
                    f'compression for history dataset found ex={e}')
            data_as_dict = parsed_data
    else:
        data_as_dict = parsed_data
    if len(data_as_dict) == 0:
        log.error(
            'empty trading history dictionary')
        return use_data

    convert_these_date_keys = [
        'date',
        'minute',
        'exp_date'
    ]

    use_include_keys = [
        'tickers',
        'version',
        'last_trade_data',
        'algo_config_dict',
        'algo_name',
        'created'
    ]
    if include_keys:
        use_include_keys = include_keys

    use_ignore_keys = []
    if ignore_keys:
        use_ignore_keys = ignore_keys

    for k in data_as_dict:
        if k in use_include_keys:
            use_data[k] = data_as_dict[k]

    all_records = []
    num_records = 0
    for ticker in data_as_dict['tickers']:
        if ticker not in use_data:
            use_data[ticker] = []
        for node in data_as_dict[ticker]:

            for ignore in use_ignore_keys:
                node.pop(ignore, None)

            all_records.append(node)
        # end for all datasets on this date to load

        num_records = len(all_records)

        if num_records:
            if verbose:
                log.info(f'found records={num_records}')
            history_df = pd.DataFrame(all_records)
            for dc in convert_these_date_keys:
                if dc in history_df:
                    history_df[dc] = pd.to_datetime(
                        history_df[dc],
                        format=ae_consts.COMMON_TICK_DATE_FORMAT)
            # end of converting all date columns
            use_data[ticker] = history_df
        else:
            log.error(
                f'did not find any records={num_records} in history dataset')
    # end for all tickers in the dataset

    return use_data
# end of prepare_history_dataset



"""
Custom Celery Task Handling
===========================

Define your own ``on_failure`` and ``on_success``
with the ``analysis_engine.work_tasks.custom_task.CustomTask`` custom
class object.

Debug values with the environment variable:

::

    export DEBUG_TASK=1

"""

import celery
import analysis_engine.consts as ae_consts
import analysis_engine.send_to_slack as slack_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


class CustomTask(celery.Task):
    """CustomTask"""

    log_label = 'custom_task'

    def build_log_label_from_args(
            self,
            args):
        """build_log_label_from_args

        :param args: list of celery args
        """
        if not args:
            return
        if len(args) > 0:
            self.log_label = f'''label=[{args[0].get(
                'label',
                args[0].get(
                    'name',
                    self.name))}]'''
        else:
            return
    # end of build_log_label_from_args

    def on_success(
            self,
            retval,
            task_id,
            args,
            kwargs):
        """on_success

        Handle custom actions when a task completes
        successfully.

        http://docs.celeryproject.org/en/latest/reference/celery.app.task.html

        :param retval: return value
        :param task_id: celery task id
        :param args: arguments passed into task
        :param kwargs: keyword arguments passed into task
        """

        self.build_log_label_from_args(
            args=args)

        if ae_consts.ev('DEBUG_TASK', '0') == '1':
            log.info(
                f'on_success {self.log_label} - retval={retval} '
                f'task_id={task_id} args={args} kwargs={kwargs}')
        else:
            log.info(f'on_success {self.log_label} - task_id={task_id}')
    # end of on_success

    def on_failure(
            self,
            exc,
            task_id,
            args,
            kwargs,
            einfo):
        """on_failure

        Handle custom actions when a task completes
        not successfully. As an example, if the task throws an
        exception, then this ``on_failure`` method can
        customize how to handle **exceptional** cases.

        http://docs.celeryproject.org/en/latest/userguide/tasks.html#task-inheritance

        :param exc: exception
        :param task_id: task id
        :param args: arguments passed into task
        :param kwargs: keyword arguments passed into task
        :param einfo: exception info
        """

        self.build_log_label_from_args(
            args=args)

        use_exc = str(exc)
        if ae_consts.ev('DEBUG_TASK', '0') == '1':
            log.error(
                f'on_failure {self.log_label} - exc={use_exc} '
                f'args={args} kwargs={kwargs}')
        else:
            log.error(f'on_failure {self.log_label} - exc={use_exc}')
        if ae_consts.ev('PROD_SLACK_ALERTS', '0') == '1':
            slack_utils.post_failure([f'on_failure {self.log_label}',
                                      f'exc={use_exc}'])
    # end of on_failure

# end of CustomTask



"""
Run an Algo

**Supported environment variables**

::

    export REDIS_ADDRESS="localhost:6379"
    export REDIS_DB="0"
    export S3_ADDRESS="localhost:9000"
    export S3_BUCKET="dev"
    export AWS_ACCESS_KEY_ID="trexaccesskey"
    export AWS_SECRET_ACCESS_KEY="trex123321"
    export AWS_DEFAULT_REGION="us-east-1"
    export S3_SECURE="0"
    export WORKER_BROKER_URL="redis://0.0.0.0:6379/13"
    export WORKER_BACKEND_URL="redis://0.0.0.0:6379/14"

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import os
import datetime
import json
import analysis_engine.consts as ae_consts
import analysis_engine.algo as base_algo
import analysis_engine.utils as ae_utils
import analysis_engine.build_algo_request as algo_utils
import analysis_engine.build_dataset_node as build_ds_node
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def run_algo(
        ticker=None,
        tickers=None,
        algo=None,  # optional derived ``analysis_engine.algo.Algo`` instance
        balance=None,     # float starting base capital
        commission=None,  # float for single trade commission for buy or sell
        start_date=None,  # string YYYY-MM-DD HH:MM:SS
        end_date=None,    # string YYYY-MM-DD HH:MM:SS
        datasets=None,    # string list of identifiers
        num_owned_dict=None,  # not supported
        cache_freq='daily',   # 'minute' not supported
        auto_fill=True,
        load_config=None,
        report_config=None,
        history_config=None,
        extract_config=None,
        use_key=None,
        extract_mode='all',
        iex_datasets=None,
        redis_enabled=True,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_key=None,
        s3_enabled=True,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        s3_key=None,
        celery_disabled=True,
        broker_url=None,
        result_backend=None,
        label=None,
        name=None,
        timeseries=None,
        trade_strategy=None,
        verbose=False,
        publish_to_slack=True,
        publish_to_s3=True,
        publish_to_redis=True,
        extract_datasets=None,
        config_file=None,
        config_dict=None,
        version=1,
        raise_on_err=True,
        **kwargs):
    """run_algo

    Run an algorithm with steps:

        1) Extract redis keys between dates
        2) Compile a data pipeline dictionary (call it ``data``)
        3) Call algorithm's ``myalgo.handle_data(data=data)``

    .. note:: If no ``algo`` is set, the
        ``analysis_engine.algo.BaseAlgo`` algorithm
        is used.

    .. note:: Please ensure Redis and Minio are running
        before trying to extract tickers

    **Stock tickers to extract**

    :param ticker: single stock ticker/symbol/ETF to extract
    :param tickers: optional - list of tickers to extract
    :param use_key: optional - extract historical key from Redis

    **Algo Configuration**

    :param algo: derived instance of ``analysis_engine.algo.Algo`` object
    :param balance: optional - float balance parameter
        can also be set on the ``algo`` object if not
        set on the args
    :param commission: float for single trade commission for
        buy or sell. can also be set on the ``algo`` objet
    :param start_date: string ``YYYY-MM-DD_HH:MM:SS`` cache value
    :param end_date: string ``YYYY-MM-DD_HH:MM:SS`` cache value
    :param dataset_types: list of strings that are ``iex`` or ``yahoo``
        datasets that are cached.
    :param cache_freq: optional - depending on if you are running data feeds
        on a ``daily`` cron (default) vs every ``minute`` (or faster)
    :param num_owned_dict: not supported yet
    :param auto_fill: optional - boolean for auto filling
        buy/sell orders for backtesting (default is
        ``True``)
    :param trading_calendar: ``trading_calendar.TradingCalendar``
        object, by default ``analysis_engine.calendars.
        always_open.AlwaysOpen`` trading calendar
        # TradingCalendar by ``TFSExchangeCalendar``
    :param config_file: path to a json file
        containing custom algorithm object
        member values (like indicator configuration and
        predict future date units ahead for a backtest)
    :param config_dict: optional - dictionary that
        can be passed to derived class implementations
        of: ``def load_from_config(config_dict=config_dict)``

    **Timeseries**

    :param timeseries: optional - string to
        set ``day`` or ``minute`` backtesting
        or live trading
        (default is ``minute``)

    **Trading Strategy**

    :param trade_strategy: optional - string to
        set the type of ``Trading Strategy``
        for backtesting or live trading
        (default is ``count``)

    **Algorithm Dataset Loading, Extracting, Reporting
    and Trading History arguments**

    :param load_config: optional - dictionary
        for setting member variables to load an
        agorithm-ready dataset from
        a file, s3 or redis
    :param report_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack
    :param history_config: optional - dictionary
        for setting member variables to publish
        an algo ``trade history`` to s3, redis, a file
        or slack
    :param extract_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack

    **(Optional) Data sources, datafeeds and datasets to gather**

    :param iex_datasets: list of strings for gathering specific `IEX
        datasets <https://iexcloud.io/>`__
        which are set as consts: ``analysis_engine.iex.consts.FETCH_*``.

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_address: Redis connection string
        format is ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_key: optional - redis key not used
        (default is ``None``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_address: Minio S3 connection string
        format ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)
    :param s3_key: optional s3 key not used
        (default is ``None``)

    **(Optional) Celery worker broker connectivity arguments**

    :param celery_disabled: bool - toggle synchronous mode or publish
        to an engine connected to the `Celery broker and backend
        <https://github.com/celery/celery#transports-and-backends>`__
        (default is ``True`` - synchronous mode without an engine
        or need for a broker or backend for Celery)
    :param broker_url: Celery broker url
        (default is ``redis://0.0.0.0:6379/13``)
    :param result_backend: Celery backend url
        (default is ``redis://0.0.0.0:6379/14``)
    :param label: tracking log label
    :param publish_to_slack: optional - boolean for
        publishing to slack (coming soon)
    :param publish_to_s3: optional - boolean for
        publishing to s3 (coming soon)
    :param publish_to_redis: optional - boolean for
        publishing to redis (coming soon)

    **(Optional) Debugging**

    :param verbose: bool - show extract warnings
        and other debug logging (default is False)
    :param raise_on_err: optional - boolean for
        unittests and developing algorithms with the
        ``analysis_engine.run_algo.run_algo`` helper.
        When set to ``True`` exceptions will
        are raised to the calling functions

    :param kwargs: keyword arguments dictionary
    """

    # dictionary structure with a list sorted on: ascending dates
    # algo_data_req[ticker][list][dataset] = pd.DataFrame
    algo_data_req = {}
    extract_requests = []
    return_algo = False  # return created algo objects for use by caller
    rec = {}
    msg = None

    use_tickers = tickers
    use_balance = balance
    use_commission = commission

    if ticker:
        use_tickers = [ticker]
    else:
        if not use_tickers:
            use_tickers = []

    # if these are not set as args, but the algo object
    # has them, use them instead:
    if algo:
        if len(use_tickers) == 0:
            use_tickers = algo.get_tickers()
        if not use_balance:
            use_balance = algo.get_balance()
        if not use_commission:
            use_commission = algo.get_commission()

    default_iex_datasets = [
        'daily',
        'minute',
        'quote',
        'stats',
        'peers',
        'news',
        'financials',
        'earnings',
        'dividends',
        'company'
    ]

    if not iex_datasets:
        iex_datasets = default_iex_datasets

    if redis_enabled:
        if not redis_address:
            redis_address = os.getenv(
                'REDIS_ADDRESS',
                'localhost:6379')
        if not redis_password:
            redis_password = os.getenv(
                'REDIS_PASSWORD',
                None)
        if not redis_db:
            redis_db = int(os.getenv(
                'REDIS_DB',
                '0'))
        if not redis_expire:
            redis_expire = os.getenv(
                'REDIS_EXPIRE',
                None)
    if s3_enabled:
        if not s3_address:
            s3_address = os.getenv(
                'S3_ADDRESS',
                'localhost:9000')
        if not s3_access_key:
            s3_access_key = os.getenv(
                'AWS_ACCESS_KEY_ID',
                'trexaccesskey')
        if not s3_secret_key:
            s3_secret_key = os.getenv(
                'AWS_SECRET_ACCESS_KEY',
                'trex123321')
        if not s3_region_name:
            s3_region_name = os.getenv(
                'AWS_DEFAULT_REGION',
                'us-east-1')
        if not s3_secure:
            s3_secure = os.getenv(
                'S3_SECURE',
                '0') == '1'
        if not s3_bucket:
            s3_bucket = os.getenv(
                'S3_BUCKET',
                'dev')
    if not broker_url:
        broker_url = os.getenv(
            'WORKER_BROKER_URL',
            'redis://0.0.0.0:6379/11')
    if not result_backend:
        result_backend = os.getenv(
            'WORKER_BACKEND_URL',
            'redis://0.0.0.0:6379/12')

    if not label:
        label = 'run-algo'

    num_tickers = len(use_tickers)
    last_close_str = ae_utils.get_last_close_str()

    if iex_datasets:
        if verbose:
            log.info(
                f'{label} - tickers={num_tickers} '
                f'iex={json.dumps(iex_datasets)}')
    else:
        if verbose:
            log.info(f'{label} - tickers={num_tickers}')

    ticker_key = use_key
    if not ticker_key:
        ticker_key = f'{ticker}_{last_close_str}'

    if not algo:
        algo = base_algo.BaseAlgo(
            ticker=None,
            tickers=use_tickers,
            balance=use_balance,
            commission=use_commission,
            config_dict=config_dict,
            name=label,
            auto_fill=auto_fill,
            timeseries=timeseries,
            trade_strategy=trade_strategy,
            publish_to_slack=publish_to_slack,
            publish_to_s3=publish_to_s3,
            publish_to_redis=publish_to_redis,
            raise_on_err=raise_on_err)
        return_algo = True
        # the algo object is stored
        # in the result at: res['rec']['algo']

    if not algo:
        msg = f'{label} - missing algo object'
        log.error(msg)
        return build_result.build_result(
                status=ae_consts.EMPTY,
                err=msg,
                rec=rec)

    if raise_on_err:
        log.debug(f'{label} - enabling algo exception raises')
        algo.raise_on_err = True

    indicator_datasets = algo.get_indicator_datasets()
    if len(indicator_datasets) == 0:
        indicator_datasets = ae_consts.BACKUP_DATASETS
        log.info(
            f'using all datasets={indicator_datasets}')

    verbose_extract = False
    if config_dict:
        verbose_extract = config_dict.get('verbose_extract', False)

    common_vals = {}
    common_vals['base_key'] = ticker_key
    common_vals['celery_disabled'] = celery_disabled
    common_vals['ticker'] = ticker
    common_vals['label'] = label
    common_vals['iex_datasets'] = iex_datasets
    common_vals['s3_enabled'] = s3_enabled
    common_vals['s3_bucket'] = s3_bucket
    common_vals['s3_address'] = s3_address
    common_vals['s3_secure'] = s3_secure
    common_vals['s3_region_name'] = s3_region_name
    common_vals['s3_access_key'] = s3_access_key
    common_vals['s3_secret_key'] = s3_secret_key
    common_vals['s3_key'] = ticker_key
    common_vals['redis_enabled'] = redis_enabled
    common_vals['redis_address'] = redis_address
    common_vals['redis_password'] = redis_password
    common_vals['redis_db'] = redis_db
    common_vals['redis_key'] = ticker_key
    common_vals['redis_expire'] = redis_expire

    use_start_date_str = start_date
    use_end_date_str = end_date
    last_close_date = ae_utils.last_close()
    end_date_val = None

    cache_freq_fmt = ae_consts.COMMON_TICK_DATE_FORMAT

    if not use_end_date_str:
        use_end_date_str = last_close_date.strftime(
            cache_freq_fmt)

    end_date_val = ae_utils.get_date_from_str(
        date_str=use_end_date_str,
        fmt=cache_freq_fmt)
    start_date_val = None

    if not use_start_date_str:
        start_date_val = end_date_val - datetime.timedelta(
            days=60)
        use_start_date_str = start_date_val.strftime(
            cache_freq_fmt)
    else:
        start_date_val = datetime.datetime.strptime(
            use_start_date_str,
            ae_consts.COMMON_TICK_DATE_FORMAT)

    total_dates = (end_date_val - start_date_val).days

    if end_date_val < start_date_val:
        msg = (
            f'{label} - invalid dates - start_date={start_date_val} is after '
            f'end_date={end_date_val}')
        raise Exception(msg)

    if verbose:
        log.info(
            f'{label} - days={total_dates} '
            f'start={use_start_date_str} '
            f'end={use_end_date_str} '
            f'datasets={indicator_datasets}')

    for ticker in use_tickers:
        req = algo_utils.build_algo_request(
            ticker=ticker,
            use_key=use_key,
            start_date=use_start_date_str,
            end_date=use_end_date_str,
            datasets=datasets,
            balance=use_balance,
            cache_freq=cache_freq,
            timeseries=timeseries,
            trade_strategy=trade_strategy,
            label=label)
        ticker_key = f'{ticker}_{last_close_str}'
        common_vals['ticker'] = ticker
        common_vals['base_key'] = ticker_key
        common_vals['redis_key'] = ticker_key
        common_vals['s3_key'] = ticker_key

        for date_key in req['extract_datasets']:
            date_req = api_requests.get_ds_dict(
                ticker=ticker,
                base_key=date_key,
                ds_id=label,
                service_dict=common_vals)
            node_date_key = date_key.replace(
                f'{ticker}_',
                '')
            extract_requests.append({
                'id': date_key,
                'ticker': ticker,
                'date_key': date_key,
                'date': node_date_key,
                'req': date_req})
    # end of for all ticker in use_tickers

    first_extract_date = None
    last_extract_date = None
    total_extract_requests = len(extract_requests)
    cur_idx = 1
    for idx, extract_node in enumerate(extract_requests):

        extract_ticker = extract_node['ticker']
        extract_date = extract_node['date']
        ds_node_id = extract_node['id']

        if not first_extract_date:
            first_extract_date = extract_date
        last_extract_date = extract_date
        perc_progress = ae_consts.get_percent_done(
            progress=cur_idx,
            total=total_extract_requests)
        percent_label = (
            f'{label} '
            f'ticker={extract_ticker} '
            f'date={extract_date} '
            f'{perc_progress} '
            f'{idx}/{total_extract_requests} '
            f'{indicator_datasets}')
        if verbose:
            log.info(
                f'extracting - {percent_label}')

        ticker_bt_data = build_ds_node.build_dataset_node(
            ticker=extract_ticker,
            date=extract_date,
            service_dict=common_vals,
            datasets=indicator_datasets,
            log_label=label,
            verbose=verbose_extract)

        if ticker not in algo_data_req:
            algo_data_req[ticker] = []

        algo_data_req[ticker].append({
            'id': ds_node_id,  # id is currently the cache key in redis
            'date': extract_date,  # used to confirm dates in asc order
            'data': ticker_bt_data
        })

        if verbose:
            log.info(
                f'extract - {percent_label} '
                f'dataset={len(algo_data_req[ticker])}')
        cur_idx += 1
    # end of for service_dict in extract_requests

    # this could be a separate celery task
    status = ae_consts.NOT_RUN
    if len(algo_data_req) == 0:
        msg = (
            f'{label} - nothing to test - no data found for '
            f'tickers={use_tickers} '
            f'between {first_extract_date} and {last_extract_date}')
        log.info(msg)
        return build_result.build_result(
            status=ae_consts.EMPTY,
            err=msg,
            rec=rec)

    # this could be a separate celery task
    try:
        if verbose:
            log.info(
                f'handle_data START - {percent_label} from '
                f'{first_extract_date} to {last_extract_date}')
        algo.handle_data(
            data=algo_data_req)
        if verbose:
            log.info(
                f'handle_data END - {percent_label} from '
                f'{first_extract_date} to {last_extract_date}')
    except Exception as e:
        a_name = algo.get_name()
        a_debug_msg = algo.get_debug_msg()
        if not a_debug_msg:
            a_debug_msg = 'debug message not set'
        a_config_dict = ae_consts.ppj(algo.config_dict)
        msg = (
            f'{percent_label} - algo={a_name} '
            f'encountered exception in handle_data tickers={use_tickers} '
            f'from {first_extract_date} to {last_extract_date} ex={e} '
            f'and failed during operation: {a_debug_msg}')
        if raise_on_err:
            if algo:
                try:
                    ind_obj = \
                        algo.get_indicator_process_last_indicator()
                    if ind_obj:
                        ind_obj_path = ind_obj.get_path_to_module()
                        ind_obj_config = ae_consts.ppj(
                            ind_obj.get_config())
                        found_error_hint = False
                        if hasattr(ind_obj.use_df, 'to_json'):
                            if len(ind_obj.use_df.index) == 0:
                                log.critical(
                                    f'indicator failure report for '
                                    f'last module: '
                                    f'{ind_obj_path} '
                                    f'indicator={ind_obj.get_name()} '
                                    f'config={ind_obj_config} '
                                    f'dataset={ind_obj.use_df.head(5)} '
                                    f'name_of_dataset={ind_obj.uses_data}')
                                log.critical(
                                    f'--------------------------------------'
                                    f'--------------------------------------')
                                log.critical(
                                    f'Please check if this indicator: '
                                    f'{ind_obj_path} '
                                    f'supports Empty Dataframes')
                                log.critical(
                                    f'--------------------------------------'
                                    f'--------------------------------------')
                                found_error_hint = True
                        # indicator error hints

                        if not found_error_hint:
                            log.critical(
                                f'indicator failure report for last module: '
                                f'{ind_obj_path} '
                                f'indicator={ind_obj.get_name()} '
                                f'config={ind_obj_config} '
                                f'dataset={ind_obj.use_df.head(5)} '
                                f'name_of_dataset={ind_obj.uses_data}')
                except Exception as f:
                    log.critical(
                        f'failed to pull indicator processor '
                        f'last indicator for debugging '
                        f'from ex={e} with parsing ex={f}')
                # end of ignoring non-supported ways of creating
                # indicator processors
            log.error(msg)
            log.error(
                f'algo failure report: '
                f'algo={a_name} handle_data() '
                f'config={a_config_dict} ')
            log.critical(
                f'algo failed during operation: {a_debug_msg}')
            raise e
        else:
            log.error(msg)
            return build_result.build_result(
                status=ae_consts.ERR,
                err=msg,
                rec=rec)
    # end of try/ex

    # this could be a separate celery task
    try:
        if verbose:
            log.info(
                f'get_result START - {percent_label} from '
                f'{first_extract_date} to {last_extract_date}')
        rec = algo.get_result()
        status = ae_consts.SUCCESS
        if verbose:
            log.info(
                f'get_result END - {percent_label} from '
                f'{first_extract_date} to {last_extract_date}')
    except Exception as e:
        msg = (
            f'{percent_label} - algo={algo.get_name()} encountered exception '
            f'in get_result tickers={use_tickers} from '
            f'{first_extract_date} to {last_extract_date} ex={e}')
        if raise_on_err:
            if algo:
                log.error(
                    f'algo={algo.get_name()} failed in get_result with '
                    f'debug_msg={algo.get_debug_msg()}')
            log.error(msg)
            raise e
        else:
            log.error(msg)
            return build_result.build_result(
                status=ae_consts.ERR,
                err=msg,
                rec=rec)
    # end of try/ex

    if return_algo:
        rec['algo'] = algo

    return build_result.build_result(
        status=status,
        err=msg,
        rec=rec)
# end of run_algo



"""
Helper for setting data in redis

Debug redis calls with:

::

    export DEBUG_REDIS=1
"""

import json
import redis
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def set_data_in_redis_key(
        label=None,
        data=None,
        client=None,
        host=None,
        port=None,
        password=None,
        db=None,
        key=None,
        expire=None,
        px=None,
        nx=False,
        xx=False,
        already_compressed=False,
        serializer='json',
        encoding='utf-8'):
    """set_data_in_redis_key

    :param label: log tracking label
    :param data: data to set in redis
    :param client: initialized redis client
    :param host: not used yet - redis host
    :param port: not used yet - redis port
    :param password: not used yet - redis password
    :param db: not used yet - redis db
    :param key: not used yet - redis key
    :param expire: redis expire
    :param px: redis px
    :param nx: redis nx
    :param xx: redis xx
    :param serializer: not used yet - support for future
        pickle objects in redis
    :param already_compressed: bool for handling
        compression to string already has happend
    :param encoding: format of the encoded key in redis
    """

    data_str = None
    encoded_data = None

    rec = {}
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    log_id = label if label else 'set-redis'

    try:
        log.debug(
            f'{log_id} serializer={serializer} encoding={encoding} '
            f'for key={key}')
        if already_compressed:
            encoded_data = data
        else:
            if serializer == 'json':
                data_str = json.dumps(data)
                encoded_data = data_str.encode(encoding)
            else:
                encoded_data = None
                err = (
                    f'{log_id} unsupported serializer={serializer} '
                    f'encoding={encoding} key={key}')
                log.error(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=rec)
                return res
        # if supported serializer

        if encoded_data:
            if ae_consts.ev('DEBUG_REDIS', '0') == '1':
                log.debug(f'{log_id} set - key={key} data={encoded_data}')

            use_client = client
            if not use_client:
                log.debug(
                    f'{log_id} set key={key} new client={host}:{port}@{db}')
                use_client = redis.Redis(
                    host=host,
                    port=port,
                    password=password,
                    db=db)
            else:
                log.debug(f'{log_id} set key={key} client')
            # create Redis client if not set

            use_client.set(
                name=key,
                value=encoded_data,
                ex=expire,
                px=px,
                nx=nx,
                xx=xx)
            res = build_result.build_result(
                status=ae_consts.SUCCESS,
                err=None,
                rec=rec)
            return res
        else:
            err = f'{log_id} no data for key={key}'
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of if have data to set
    except Exception as e:
        err = (
            f'{log_id} failed - redis set from data={str(data)[0:200]} '
            f'encoded_data={str(encoded_data)[0:200]} '
            f'key={key} ex={e}')
        log.error(err)
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex for setting redis data

    return res
# end of set_data_in_redis_key
